#!/usr/bin/env python2.6
'''%prog [options] GENE-REF INPUT [OUTPUT]
GENE-REF is the file of gene information.
INPUT is a BED formatted file of peaks.
OUTPUT is an optional output file name.'''
#Last modified on Jan 14, 2010
#@author: Brian Kennedy

import SARx
import sys
from optparse import OptionParser
from bisect import bisect_left
from itertools import groupby, chain, izip_longest

def unique(iterable, key=None):
	for group in groupby(iterable, key):
		yield group[1].next()

def grouper(n, iterable, fillvalue=None):
	"grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
	args = [iter(iterable)] * n
	return izip_longest(fillvalue=fillvalue, *args)

def parseGeneLine(line):
	fields = line.strip().split()
	if fields[3] == '+':
		fivePrime, threePrime = int(fields[4]), int(fields[5])
	else:
		fivePrime, threePrime = int(fields[5]), int(fields[4])
	return ("chr%s" % fields[2], fivePrime, threePrime, fields[3], #chr 5' 3' strand 
		fields[0], fields[6], fields[7]) #RefseqID GeneSymbol GenbankID

def parseExonLine(line):
	"a generator that returns all of the exons as individual 'gene' elements."
	fields = line.strip().split()
	exons = list(grouper(2, fields[8:]))
	geneFivePrime = int(exons[0][0])
	geneThreePrime = int(exons[-1][1])
	for num, (fivePrime, threePrime) in enumerate(exons, 1):
		yield ["chr%s" % fields[2], int(fivePrime), int(threePrime), fields[3], #chr exon-5' exon-3' strand 
			geneFivePrime, geneThreePrime, fields[0], fields[6], fields[7], num, len(exons), fields] #gene-5' gene-3' RefseqID Gene Symbol GenbankID

def geneInfo(info):
	genes = SARx.keyByIndex(info)
	peak = (0, 0, 0)
	for k,v in genes.items():
		genes[k] = {}
		for distance in distances:
			genes[k][distance] = {"data":sorted(v, key=lambda x: abs(distance(peak, x)))}
			genes[k][distance]["data"] = list(unique(genes[k][distance]["data"], key=lambda x: abs(distance(peak,x))))
			genes[k][distance]["coordinates"] = \
				[abs(distance(peak, gene)) for gene in genes[k][distance]["data"]]
	return genes

def checkRegion(peak, gene, region):
	distance = region["distance"](peak, gene)
	if isinstance(distance, bool):
		return (distance, distances[0](peak, gene))
	if "min" not in region:
		region["min"] = 0
	return (region["min"] <= distance < region["max"], distance)

def findRegion(peak, gene, regions):
	for n, region in enumerate(regions):
		valid, distance = checkRegion(peak, gene, region)
		if valid: return n, distance

def findGene(peak, geneInfo, distance):
	mid = (peak[1]+peak[2]) / 2
	index = bisect_left(geneInfo[distance]["coordinates"], mid)
	if index > 0:
		if index == len(geneInfo[distance]["data"]):
			targets = (index-1,)
		else:
			targets = (index-1,index)
	else:
		targets = (0,)
	try:
		return min([geneInfo[distance]["data"][i] for i in targets],
		key=lambda x: abs(distance(peak, x)))
	except IndexError:
		print("index %s targets %s" %(index, targets))

def findOptimalRegion(peak, geneInfo):
	genes = [findGene(peak, geneInfo, d) for d in distances]
	optimalGene = min(genes, key=lambda x: findRegion(peak, x, regions))
	optimalRegion, distance = findRegion(peak, optimalGene, regions)
	if regions[optimalRegion]["distance"] == toThreePrime:
		distance *= -1
	return (optimalRegion, optimalGene, distance)

def toFivePrime(peak, gene):
	mid = (peak[1]+peak[2]) / 2
	return (gene[1] - mid) if gene[3]=='+' else (mid - gene[1])

def toThreePrime(peak, gene):
	mid = (peak[1]+peak[2]) / 2
	return (mid - gene[2]) if gene[3]=='+' else (gene[2] - mid)

def isIntragenic(peak, gene):
	mid = (peak[1]+peak[2]) / 2
	return gene[1] < mid < gene[2] if gene[3]=='+' else gene[1] > mid > gene[2]

def isGeneDesert(peak, gene):
	return True

def measureOverlap(peak, element):
	"counts the number of overlapping nucleotides between peak and exon"
	exon = (element[4], element[5]) if element[3] == '+' else (element[5], element[4])
	if peak[1] <= exon[0] <= peak[2]:
		return min(peak[2] - exon[0], exon[1] - exon[0])
	elif peak[1] <= exon[1] <= peak[2]:
		return min(exon[1] - peak[1], exon[1] - exon[0])
	elif exon[0] <= peak[1] and peak[2] <= exon[1]:
		return peak[2] - peak[1]
	else:
		return 0

def alignPeaks(peakData, geneDict):
	for peak in peakData:
		try:
			region, gene, distance = findOptimalRegion(peak, geneDict[peak[0]])
		except KeyError:
			sys.stderr.write("%s has no genes in %s.\n" % (args[0], peak[0]))
		yield peak, region, gene, -1 * distance
	
def getGeneAlignments(geneFile, peakData):
	geneDict = geneInfo(parseGeneLine(line) for line in geneFile)
	return alignPeaks(peakData, geneDict)
	
def getExonAlignments(geneFile, peakData):
	geneDict = geneInfo(chain.from_iterable(parseExonLine(line) for line in geneFile))
	for peak, exonRegion, exon, exonDistance in alignPeaks(peakData, geneDict):
		gene = exon[4:6]
		exon[4:6] = exon[1:3]
		exon[1:3] = gene
		geneRegion, geneDistance = findRegion(peak, exon, geneRegions)
		if geneRegions[geneRegion]["distance"] == toThreePrime:
			geneDistance *= -1
		yield (peak, geneRegion, exon, -1 * geneDistance)

def formatExon(peak, region, gene, distance):
	overlap = measureOverlap(peak, gene)
	exonFields = gene[11]
	geneData = (gene[0], gene[1], gene[2], gene[3])
	peakMid = (peak[1]+peak[2]) / 2
	exonMid = (gene[4]+gene[5]) / 2
	if gene[3] == '+' and peakMid > exonMid or gene[3] == '-' and peakMid < exonMid:
		if gene[9]+1<=gene[10]:
			exonLabel = "exon%d_intron%d" % (gene[9], gene[9])
		else:
			exonLabel = "exon%d_3" % (gene[9])
	else:
		if gene[9]-1>=1:
			exonLabel = "intron%d_exon%d" % (gene[9]-1, gene[9])
		else:
			exonLabel = "exon%d_5" % (gene[9])
	geneRegionName = geneRegions[region]["name"]
	label = geneRegionName if geneRegionName != "Intragenic" else exonLabel
	return (peak, label, distance, overlap, exonFields)

def formatGene(peak, region, gene, distance):
	name = regions[region]["name"]
	geneFields = gene[1:]
	return (peak, name, distance, geneFields)
			
# These regions are listed in their order of precedence.
# distance assumes upstream of 5' and down stream of 3', so in both cases, 
# an intragenic location will be negative. e.g. the distance to 5' of a peak at 1200bp
#  from a gene with 5' at 1700bp will be 500.  The distance to 3' of a peak at 1400bp
# from a gene with a 3' at 1100bp will be 300bp.
geneRegions = (
	{"name":"5_TSS", "distance":toFivePrime, "min":-1000, "max":1000},
	{"name":"3_Core", "distance":toThreePrime, "max":2000},
	{"name":"5_Proximal", "distance":toFivePrime, "max":10000},
	{"name":"3_Proximal", "distance":toThreePrime, "max":10000},
	{"name":"Intragenic", "distance":isIntragenic},
	{"name":"5_Distal", "distance":toFivePrime,"max":100000},
	{"name":"3_Distal", "distance":toThreePrime, "max":100000},
	{"name":"Gene_Desert", "distance":isGeneDesert},
	)
exonRegions = (
	{"name":"5_Exon", "distance":toFivePrime, "max":1000},
	{"name":"3_Exon", "distance":toThreePrime, "max":1000},
	{"name":"Inside_Exon", "distance":isIntragenic},
	{"name":"Exon_Desert", "distance":isGeneDesert},
)
distances = (toFivePrime, toThreePrime)
	
if __name__ == "__main__":
	parser = OptionParser(usage=__doc__, version="%prog 0.4")
	parser.add_option("-e", "--exon", dest="useExons",
		action="store_true", default=False,
		help="Calculate distances to exons rather than genes and output the peak/exon overlap.\
This requires an exon reference file instead of gene.")
	parser.add_option("-a", "--all-fields", dest="allFields", 
		action="store_true", default=False,
		help="Print all of the input's fields in the output.")
	parser.add_option("-5", "--five-prime", dest="fivePrimeDistance", 
		action="store_true", default=False,
		help="Calculate distances as normal, but output 5' distance.")
	parser.add_option("-3", "--three-prime", dest="threePrimeDistance", 
		action="store_true", default=False,
		help="Calculate distances as normal, but output 3' distance.")

	(options, args) = parser.parse_args()

	if len(args) < 2:
		parser.print_help()
		exit(1)
	if len(args) < 3:
		output = sys.stdout
	else:
		output = open(args[2], "w", 4096)
		
	geneFile = SARx.openAnything(args[0])
	peakFile = SARx.openAnything(args[1])
	peakData = SARx.loadBed(peakFile, all=options.allFields)
	
	if options.useExons:
		regions = exonRegions
		align = getExonAlignments
		format = formatExon
	else:
		regions = geneRegions
		align = getGeneAlignments
		format = formatGene
		
	for peak, region, gene, distance in align(geneFile, peakData):
		if options.fivePrimeDistance or options.threePrimeDistance:
			distance = []
			if options.fivePrimeDistance:
				distance.append(-1 * toFivePrime(peak, gene))
			if options.threePrimeDistance:
				distance.append(toThreePrime(peak,gene))
		fields = format(peak, region, gene, distance)
		output.write("%s\n" % SARx.tabFormat(fields))
